#ifndef _MCK_MCKMBLOCK_H
#define _MCK_MCKMBLOCK_H
#ifdef __KERNEL__

/*
 * Logical memory blocks.
 *
 * Copyright (C) 2001 Peter Bergner, IBM Corp.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/init.h>
#include <linux/mm.h>

#define MCK_MAX_NUMNODES 16
#define INIT_MCKMBLOCK_REGIONS	128

#define sw_memblock_is_region_reserved(base, size) mckmblock_is_region_reserved(base, size)
#define sw_memblock_reserve(base, size) mckmblock_reserve(base, size)
#define sw__memblock_alloc_base(size, alignment, limit) __mckmblock_alloc_base(size, alignment, limit)
#define sw_for_each_memblock(memblock_type, region) for_each_mckmblock(memblock_type, region)
#define sw_memblock_region_memory_end_pfn(reg) mckmblock_region_memory_end_pfn(reg)
#define sw_memblock_region_memory_base_pfn(reg) mckmblock_region_memory_base_pfn(reg)
#define sw_memblock_region mckmblock_region

struct mckmblock_region {
	phys_addr_t base;
	phys_addr_t size;
	int nid;
};

struct mckmblock_type {
	unsigned long cnt;	/* number of regions */
	unsigned long max;	/* size of the allocated array */
	phys_addr_t total_size;	/* size of all regions */
	struct mckmblock_region *regions;
};

struct mckmblock {
	phys_addr_t current_limit;
	struct mckmblock_type memory;
	struct mckmblock_type reserved;
};

extern struct mckmblock mckmblock;
extern int mckmblock_debug;

#define mckmblock_dbg(fmt, ...) \
	if (mckmblock_debug)	\
		printk(KERN_INFO pr_fmt(fmt), ##__VA_ARGS__)

phys_addr_t mckmblock_find_in_range_node(phys_addr_t start, phys_addr_t end,
				phys_addr_t size, phys_addr_t align, int nid);
phys_addr_t mckmblock_find_in_range(phys_addr_t start, phys_addr_t end,
				   phys_addr_t size, phys_addr_t align);
phys_addr_t get_allocated_mckmblock_reserved_regions_info(phys_addr_t *addr);
void mckmblock_allow_resize(void);
int mckmblock_add_node(phys_addr_t base, phys_addr_t size, int nid);
int mckmblock_add(phys_addr_t base, phys_addr_t size);
int mckmblock_remove(phys_addr_t base, phys_addr_t size);
int mckmblock_free(phys_addr_t base, phys_addr_t size);
int mckmblock_reserve(phys_addr_t base, phys_addr_t size);
void mckmblock_trim_memory(phys_addr_t align);

void __next_mckmem_pfn_range(int *idx, int nid, unsigned long *out_start_pfn,
			  unsigned long *out_end_pfn, int *out_nid);

/**
 * for_each_mem_pfn_range - early memory pfn range iterator
 * @i: an integer used as loop variable
 * @nid: node selector, %MAX_NUMNODES for all nodes
 * @p_start: ptr to ulong for start pfn of the range, can be %NULL
 * @p_end: ptr to ulong for end pfn of the range, can be %NULL
 * @p_nid: ptr to int for nid of the range, can be %NULL
 *
 * Walks over configured memory ranges.
 */
#ifndef for_each_mem_pfn_range
#define for_each_mem_pfn_range(i, nid, p_start, p_end, p_nid)		\
	for (i = -1, __next_mckmem_pfn_range(&i, nid, p_start, p_end, p_nid); \
	     i >= 0; __next_mckmem_pfn_range(&i, nid, p_start, p_end, p_nid))
#endif

void __next_free_mckmem_range(u64 *idx, int nid, phys_addr_t *out_start,
			   phys_addr_t *out_end, int *out_nid);

/**
 * for_each_free_mem_range - iterate through free mckmblock areas
 * @i: u64 used as loop variable
 * @nid: node selector, %MAX_NUMNODES for all nodes
 * @p_start: ptr to phys_addr_t for start address of the range, can be %NULL
 * @p_end: ptr to phys_addr_t for end address of the range, can be %NULL
 * @p_nid: ptr to int for nid of the range, can be %NULL
 *
 * Walks over free (memory && !reserved) areas of mckmblock.  Available as
 * soon as mckmblock is initialized.
 */
#ifndef for_each_free_mem_range
#define for_each_free_mem_range(i, nid, p_start, p_end, p_nid)		\
	for (i = 0,							\
	     __next_free_mckmem_range(&i, nid, p_start, p_end, p_nid);	\
	     i != (u64)ULLONG_MAX;					\
	     __next_free_mckmem_range(&i, nid, p_start, p_end, p_nid))
#endif

void __next_free_mckmem_range_rev(u64 *idx, int nid, phys_addr_t *out_start,
			       phys_addr_t *out_end, int *out_nid);

/**
 * for_each_free_mem_range_reverse - rev-iterate through free mckmblock areas
 * @i: u64 used as loop variable
 * @nid: node selector, %MAX_NUMNODES for all nodes
 * @p_start: ptr to phys_addr_t for start address of the range, can be %NULL
 * @p_end: ptr to phys_addr_t for end address of the range, can be %NULL
 * @p_nid: ptr to int for nid of the range, can be %NULL
 *
 * Walks over free (memory && !reserved) areas of mckmblock in reverse
 * order.  Available as soon as mckmblock is initialized.
 */
#ifndef for_each_free_mem_range_reverse
#define for_each_free_mem_range_reverse(i, nid, p_start, p_end, p_nid)	\
	for (i = (u64)ULLONG_MAX,					\
	     __next_free_mckmem_range_rev(&i, nid, p_start, p_end, p_nid);	\
	     i != (u64)ULLONG_MAX;					\
	     __next_free_mckmem_range_rev(&i, nid, p_start, p_end, p_nid))
#endif

int mckmblock_set_node(phys_addr_t base, phys_addr_t size, int nid);

static inline void mckmblock_set_region_node(struct mckmblock_region *r, int nid)
{
	r->nid = nid;
}

static inline int mckmblock_get_region_node(const struct mckmblock_region *r)
{
	return r->nid;
}
phys_addr_t mckmblock_alloc_nid_bottom_up(phys_addr_t size, phys_addr_t align, int nid);
phys_addr_t mckmblock_alloc_nid(phys_addr_t size, phys_addr_t align, int nid);
phys_addr_t mckmblock_alloc_try_nid(phys_addr_t size, phys_addr_t align, int nid);

phys_addr_t mckmblock_alloc(phys_addr_t size, phys_addr_t align);

/* Flags for mckmblock_alloc_base() amd __mckmblock_alloc_base() */
#define MCKMBLOCK_ALLOC_ANYWHERE	(~(phys_addr_t)0)
#define MCKMBLOCK_ALLOC_ACCESSIBLE	0

phys_addr_t mckmblock_alloc_base_bottom_up(phys_addr_t size, phys_addr_t align,
				phys_addr_t max_addr);
phys_addr_t mckmblock_alloc_base(phys_addr_t size, phys_addr_t align,
				phys_addr_t max_addr);
phys_addr_t __mckmblock_alloc_base(phys_addr_t size, phys_addr_t align,
				  phys_addr_t max_addr);
phys_addr_t mckmblock_phys_mem_size(void);
phys_addr_t mckmblock_start_of_DRAM(void);
phys_addr_t mckmblock_end_of_DRAM(void);
void mckmblock_enforce_memory_limit(phys_addr_t memory_limit);
int mckmblock_is_memory(phys_addr_t addr);
int mckmblock_is_region_memory(phys_addr_t base, phys_addr_t size);
int mckmblock_is_reserved(phys_addr_t addr);
int mckmblock_is_region_reserved(phys_addr_t base, phys_addr_t size);

extern void __mckmblock_dump_all(void);

static inline void mckmblock_dump_all(void)
{
	if (mckmblock_debug)
		__mckmblock_dump_all();
}

/**
 * mckmblock_set_current_limit - Set the current allocation limit to allow
 *                         limiting allocations to what is currently
 *                         accessible during boot
 * @limit: New limit value (physical address)
 */
void mckmblock_set_current_limit(phys_addr_t limit);


/*
 * pfn conversion functions
 *
 * While the memory mckmblocks should always be page aligned, the reserved
 * mckmblocks may not be. This accessor attempt to provide a very clear
 * idea of what they return for such non aligned mckmblocks.
 */

/**
 * mckmblock_region_memory_base_pfn - Return the lowest pfn intersecting with the memory region
 * @reg: mckmblock_region structure
 */
static inline unsigned long mckmblock_region_memory_base_pfn(const struct mckmblock_region *reg)
{
	return PFN_UP(reg->base);
}

/**
 * mckmblock_region_memory_end_pfn - Return the end_pfn this region
 * @reg: mckmblock_region structure
 */
static inline unsigned long mckmblock_region_memory_end_pfn(const struct mckmblock_region *reg)
{
	return PFN_DOWN(reg->base + reg->size);
}

/**
 * mckmblock_region_reserved_base_pfn - Return the lowest pfn intersecting with the reserved region
 * @reg: mckmblock_region structure
 */
static inline unsigned long mckmblock_region_reserved_base_pfn(const struct mckmblock_region *reg)
{
	return PFN_DOWN(reg->base);
}

/**
 * mckmblock_region_reserved_end_pfn - Return the end_pfn this region
 * @reg: mckmblock_region structure
 */
static inline unsigned long mckmblock_region_reserved_end_pfn(const struct mckmblock_region *reg)
{
	return PFN_UP(reg->base + reg->size);
}

#define for_each_mckmblock(mckmblock_type, region)					\
	for (region = mckmblock.mckmblock_type.regions;				\
	     region < (mckmblock.mckmblock_type.regions + mckmblock.mckmblock_type.cnt);	\
	     region++)


int mckmblock_internal_test(void);
#endif /* __KERNEL__ */

#define MCKMBLOCKIO 0xAF

/*
 * ioctls for /dev/mck fds:
 */
#define MCKMBLOCK_GET_API_VERSION       _IO(MCKMBLOCKIO,   0x00)
#define MCKMBLOCK_ALLOC                 _IOWR(MCKMBLOCKIO, 0x02, void *)
#define MCKMBLOCK_FREE                  _IOWR(MCKMBLOCKIO, 0x03, unsigned long)

#endif /* _MCK_MCKMBLOCK_H */
